import atexit
import logging
import numbers

from track.const import ApiPaths
from track.consumer import Consumer
from track.request import post
from track.utils import require, stringify

try:
    import queue
except ImportError:
    import Queue as queue

ID_TYPES = (numbers.Number, str)


class Client(object):
    logger = logging.getLogger('interakt')

    def __init__(self, write_key=None, host=None, debug=False,
                 sync_mode=True, timeout=10, max_queue_size=10000,
                 on_error=None, max_retries=3, flush_interval=0.5):
        """Create a new interakt client"""
        require('write_key', write_key, str)

        self.queue = queue.Queue(maxsize=max_queue_size)
        self.write_key = write_key
        self.debug = debug
        self.host = host
        self.sync_mode = sync_mode
        self.timeout = timeout
        self.on_error = on_error
        if debug:
            self.logger.setLevel(logging.DEBUG)

        if sync_mode:
            self.consumer = None
        else:
            atexit.register(self.join)
            self.consumer = Consumer(
                queue=self.queue, write_key=write_key,
                host=host, on_error=on_error, retries=max_retries,
                timeout=timeout, flush_interval=flush_interval
            )
            self.consumer.start()

    def identify(self, user_id=None, country_code='+91', phone_number=None, traits={}):
        """Tie a user to their actions and record traits about them."""
        if not user_id and not phone_number:
            raise AssertionError("Either user_id or phone_number is required")
        if user_id:
            require('user_id', user_id, ID_TYPES)
        if phone_number:
            require('phone_number', phone_number, str)
        require('traits', traits, dict)
        body = {
            'userId': stringify(val=user_id),
            'countryCode': country_code,
            'phoneNumber': phone_number,
            'traits': traits
        }
        return self.__queue_request(path=ApiPaths.Identify.value, body=body)

    def event(self, user_id=None, event=None, traits={}):
        """To record user events"""
        traits = traits or {}
        require('user_id', user_id, ID_TYPES)
        require('traits', traits, dict)
        require('event', event, str)
        body = {
            'userId': stringify(val=user_id),
            'event': event,
            'traits': traits
        }
        return self.__queue_request(path=ApiPaths.Event.value, body=body)

    def flush(self):
        """Forces a flush from the internal queue to the server"""
        queue = self.queue
        size = queue.qsize()
        queue.join()
        # Note that this message may not be precise, because of threading.
        self.logger.debug('Successfully flushed about %s items.', size)

    def join(self):
        """Ends the consumer thread once the queue is empty.
        Blocks execution until finished
        """
        self.consumer.pause()
        try:
            self.consumer.join()
        except RuntimeError:
            # consumer thread has not started
            pass

    def shutdown(self):
        """Flush all messages and cleanly shutdown the client"""
        self.flush()
        self.join()

    def __queue_request(self, path, body):
        # Directly call api in sync mode and return response
        if self.sync_mode:
            return post(self.write_key, host=self.host, path=path, body=body, timeout=self.timeout)

        queue_msg = {
            'path': path,
            'body': body
        }
        try:
            self.queue.put(queue_msg, block=False)
            self.logger.debug(f'Enqueued msg for {path}')
            return True, queue_msg
        except queue.Full:
            self.logger.warning('track-python queue is full')
            return False, queue_msg



import logging

from requests import sessions
from requests.auth import HTTPBasicAuth

from track.utils import remove_trailing_slash
from track.version import VERSION
from track.const import DEFAULT_HOST

_session = sessions.session()
logger = logging.getLogger('interakt')


def post(write_key, host=None, path=None, body=None, timeout=10):
    """Post the msg to the API"""
    auth = HTTPBasicAuth(username=write_key, password="")
    headers = {
        'Content-Type': 'application/json',
        'User-Agent': f'interakt-track-python/{VERSION}'
    }
    url = remove_trailing_slash(host or DEFAULT_HOST) + path
    logger.debug(f'Making request: {body}')
    response = _session.post(url=url, headers=headers,
                             auth=auth, json=body, timeout=timeout)
    if response.status_code == 200:
        logger.debug("Data uploaded successfully")
        return response

    try:
        payload = response.json()
        logger.debug(f'Received response: {payload}')
        raise APIError(payload.get("result"),
                       response.status_code, payload.get("message"))
    except ValueError:
        raise APIError('Unknown', response.status_code, response.text)


class APIError(Exception):

    def __init__(self, status, code, message):
        self.message = message
        self.status = status
        self.code = code

    def __str__(self):
        msg = "[interakt-track] StatusCode({0}): {1} (Success={2})"
        return msg.format(self.code, self.message, self.status)



import logging
from threading import Thread
from track.request import post, APIError
import backoff

try:
    from queue import Empty
except ImportError:
    from Queue import Empty


class Consumer(Thread):
    """Consumes the messages from the client's queue."""
    logger = logging.getLogger('interakt')

    def __init__(self, queue, write_key, host=None,
                 on_error=None, retries=3, timeout=10,
                 flush_interval=0.5):
        Thread.__init__(self)
        self.running = True
        self.queue = queue
        self.write_key = write_key
        self.host = host
        self.on_error = on_error
        self.retries = retries
        self.timeout = timeout
        self.flush_interval = flush_interval

    def run(self):
        """Runs the consumer."""
        self.logger.debug('consumer is running...')
        while self.running:
            self.upload()

        self.logger.debug('consumer exited.')

    def pause(self):
        """Pause the consumer."""
        self.running = False

    def upload(self):
        queue = self.queue
        queue_msg = None
        try:
            queue_msg = queue.get(block=True, timeout=self.flush_interval)
        except Empty:
            self.logger.debug("queue is empty now")

        if not queue_msg:
            self.logger.debug("Nothing left in queue exiting")
            return False
        try:
            self.request(queue_msg=queue_msg)
            success = True
        except Exception as e:
            self.logger.error(f"Error uploading: {e}")
            success = False
            if self.on_error:
                self.on_error(e, queue_msg)
        finally:
            self.queue.task_done()
        return success

    def request(self, queue_msg):
        """Attempt to upload the queue_msg and retry before raising an error """

        def fatal_exception(exc):
            if isinstance(exc, APIError):
                # retry on server errors and client errors
                # with 429 status code (rate limited),
                # don't retry on other client errors
                return (400 <= exc.status < 500) and exc.status != 429
            else:
                # retry on all other errors (eg. network)
                return False

        @backoff.on_exception(
            backoff.expo,
            Exception,
            max_tries=self.retries + 1,
            giveup=fatal_exception)
        def send_request():
            post(write_key=self.write_key, host=self.host,
                 path=queue_msg['path'], body=queue_msg['body'], timeout=self.timeout)

        send_request()



from track.version import VERSION
from track.client import Client
__version__ = VERSION

"""Settings"""
write_key = None
default_client = None
host = None
sync_mode = False
debug = False
timeout = 10
max_retries = 3
on_error = None
max_queue_size = 10000


def identify(user_id=None, country_code='+91', phone_number=None, traits={}):
    """Send an identify call for customer"""
    return _proxy('identify', user_id=user_id, country_code=country_code,
                  phone_number=phone_number, traits=traits)


def event(user_id=None, event=None, traits={}):
    """Send an event track call."""
    return _proxy('event', user_id=user_id, event=event, traits=traits)


def flush():
    """Tell the client to flush."""
    _proxy('flush')


def join():
    """Block program until the client clears the queue"""
    _proxy('join')


def shutdown():
    """Flush all messages and cleanly shutdown the client"""
    _proxy('shutdown')


def _proxy(method, *args, **kwargs):
    """Create an analytics client if one doesn't exist and send to it."""
    global default_client
    if not default_client:
        default_client = Client(
            write_key=write_key,
            host=host,
            debug=debug,
            sync_mode=sync_mode,
            timeout=timeout,
            max_queue_size=max_queue_size,
            on_error=on_error,
            max_retries=max_retries
        )

    fn = getattr(default_client, method)
    return fn(*args, **kwargs)



import logging

logger = logging.getLogger('interakt')


def require(name, field, data_type):
    """Require that the named `field` has the right `data_type`"""
    if not isinstance(field, data_type):
        msg = '{0} must have {1}, got: {2}'.format(name, data_type, field)
        raise AssertionError(msg)


def remove_trailing_slash(host):
    if host.endswith('/'):
        return host[:-1]
    return host


def stringify(val):
    if val is None:
        return None
    if isinstance(val, str):
        return val
    return str(val)




import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# Don't import interakt-track-python module here, since deps may not be installed
sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'track'))
from track.version import VERSION  # noqa


# TODO: update long description in README.md
with open(file='README.md', mode='r', encoding='utf-8') as f:
    readme = f.read()

install_requires = [
    "requests>=2.20,<3.0",
    "backoff==1.10.0"
]

tests_require = [
    "mock==4.0.2",
    "pylint==2.6.0",
    "flake8==3.8.3",
    "coverage==5.2.1"
]


setup(
    name='interakt-track-python',
    packages=['track', 'track.tests'],
    version=VERSION,
    url='https://github.com/interakt/track-python',
    author="Amar Jaiswal",
    author_email="amar.j@cawstudios.com",
    maintainer="interakt.ai",
    license='MIT License',
    description='The easy way to integrate track apis for interakt',
    keywords=['INTERAKT', 'KIWI'],
    install_requires=install_requires,
    extras_require={'test': tests_require},
    long_description=readme,
    long_description_content_type='text/markdown',
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)
